@@ -42,6 +42,8 @@
 #include <telepathy-glib/util.h>
 
 #include <gst/farsight/fs-conference-iface.h>
+#include <gst/farsight/fs-rtp.h>
+#include <gst/farsight/fs-utils.h>
 
 #include "stream.h"
 #include "stream-priv.h"
@@ -73,6 +75,11 @@ G_DEFINE_TYPE (TfStream, tf_stream, G_TYPE_OBJECT);
 
 static TpMediaStreamError fserrorno_to_tperrorno (FsError fserror);
 
+struct DtmfEvent {
+  gint codec_id;
+  guint event_id;
+};
+
 struct _TfStreamPrivate
 {
   TfChannel *channel;
@@ -97,9 +104,15 @@ struct _TfStreamPrivate
 
   gboolean send_local_codecs;
   gboolean send_supported_codecs;
+  gboolean ready_called;
 
   guint tos;
 
+  GHashTable *feedback_messages;
+  GPtrArray *header_extensions;
+
+  gboolean block_ready;
+
   GStaticMutex mutex;
   guint idle_connected_id; /* Protected by mutex */
   gboolean disposed; /* Protected by mutex */
@@ -107,6 +120,10 @@ struct _TfStreamPrivate
   TpMediaStreamState current_state;
 
   NewStreamCreatedCb *new_stream_created_cb;
+
+  GQueue events_to_send;
+
+  gint sending_telephony_event;
 };
 
 enum
@@ -137,7 +154,8 @@ enum
   PROP_SINK_PAD,
   PROP_LOCAL_PREFERENCES,
   PROP_TOS,
-  PROP_RESOURCES
+  PROP_RESOURCES,
+  PROP_BLOCK_READY
 };
 
 static void get_all_properties_cb (TpProxy *proxy,
@@ -180,12 +198,24 @@ static void set_stream_sending (TpMediaStreamHandler *proxy, gboolean play,
 static void start_telephony_event (TpMediaStreamHandler *proxy, guchar event,
     gpointer user_data, GObject *object);
 
+static void start_named_telephony_event (TpMediaStreamHandler *proxy,
+    guchar event, guint codecid, gpointer user_data, GObject *object);
+
+static void start_sound_telephony_event (TpMediaStreamHandler *proxy,
+    guchar event, gpointer user_data, GObject *object);
+
 static void stop_telephony_event (TpMediaStreamHandler *proxy,
     gpointer user_data, GObject *object);
 
 static void stream_close (TpMediaStreamHandler *proxy,
     gpointer user_data, GObject *object);
 
+static void set_remote_feedback_messages (TpMediaStreamHandler *proxy,
+    GHashTable *messages, gpointer user_data, GObject *object);
+
+static void set_remote_header_extensions (TpMediaStreamHandler *proxy,
+    const GPtrArray *header_extensions, gpointer user_data, GObject *object);
+
 static void invalidated_cb (TpMediaStreamHandler *proxy,
     guint domain, gint code, gchar *message, gpointer user_data);
 
@@ -225,6 +255,9 @@ tf_stream_init (TfStream *self)
   g_static_mutex_init (&priv->mutex);
   priv->has_resource = TP_MEDIA_STREAM_DIRECTION_NONE;
   priv->current_state = TP_MEDIA_STREAM_STATE_DISCONNECTED;
+  priv->sending_telephony_event = -1;
+
+  g_queue_init (&priv->events_to_send);
 }
 
 static void
@@ -284,6 +317,9 @@ tf_stream_get_property (GObject    *object,
     case PROP_RESOURCES:
       g_value_set_uint (value, self->priv->has_resource);
       break;
+    case PROP_BLOCK_READY:
+      g_value_set_boolean (value, self->priv->block_ready);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -336,6 +372,20 @@ tf_stream_set_property (GObject      *object,
       if (self->priv->fs_session)
         g_object_set_property (G_OBJECT (self->priv->fs_session), "tos", value);
       break;
+    case PROP_BLOCK_READY:
+      if (self->priv->ready_called)
+        {
+          WARNING (self,
+              "Trying to block the Ready() call after it has happenned");
+        }
+      else
+        {
+          gboolean old_block_ready = self->priv->block_ready;
+          self->priv->block_ready = g_value_get_boolean (value);
+          if (old_block_ready && !self->priv->block_ready)
+            _tf_stream_try_sending_codecs (self);
+        }
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -373,6 +423,7 @@ tf_stream_dispose (GObject *object)
 {
   TfStream *stream = TF_STREAM (object);
   TfStreamPrivate *priv = stream->priv;
+  gpointer data;
 
   TF_STREAM_LOCK (stream);
   if (stream->priv->idle_connected_id)
@@ -399,6 +450,7 @@ tf_stream_dispose (GObject *object)
       tf_stream_free_resource (stream,
           TP_MEDIA_STREAM_DIRECTION_RECEIVE);
 
+      g_object_run_dispose (G_OBJECT (priv->fs_stream));
       g_object_unref (priv->fs_stream);
 
       tf_stream_free_resource (stream,
@@ -409,6 +461,7 @@ tf_stream_dispose (GObject *object)
 
   if (priv->fs_session)
     {
+      g_object_run_dispose (G_OBJECT (priv->fs_session));
       g_object_unref (priv->fs_session);
       priv->fs_session = NULL;
     }
@@ -437,6 +490,19 @@ tf_stream_dispose (GObject *object)
       priv->last_sent_codecs = NULL;
     }
 
+  if (priv->feedback_messages)
+    g_boxed_free (TP_HASH_TYPE_RTCP_FEEDBACK_MESSAGE_MAP,
+        priv->feedback_messages);
+  priv->feedback_messages = NULL;
+
+  if (priv->header_extensions)
+    g_boxed_free (TP_ARRAY_TYPE_RTP_HEADER_EXTENSIONS_LIST,
+        priv->header_extensions);
+  priv->header_extensions = NULL;
+
+  while ((data = g_queue_pop_head (&priv->events_to_send)))
+    g_slice_free (struct DtmfEvent, data);
+
   fs_candidate_list_destroy (priv->local_candidates);
   priv->local_candidates = NULL;
 
@@ -581,6 +647,13 @@ tf_stream_class_init (TfStreamClass *klass)
           TP_MEDIA_STREAM_DIRECTION_NONE,
           G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
 
+  g_object_class_install_property (object_class, PROP_BLOCK_READY,
+      g_param_spec_boolean ("block-ready",
+          "Blocks calling Ready on the StreamHandler",
+          "Blocks calling Ready on the StreamHandler",
+          FALSE,
+          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
+
   /**
    * TfStream::closed:
    * @stream: the stream that has been closed
@@ -697,6 +770,7 @@ get_all_properties_cb (TpProxy *proxy,
   gboolean valid = FALSE;
   guint i;
   gboolean do_controlling = FALSE;
+  GList *rtp_header_extensions;
 
   if (dbus_error &&
       !(dbus_error->domain == DBUS_GERROR &&
@@ -734,12 +808,24 @@ get_all_properties_cb (TpProxy *proxy,
   tp_cli_media_stream_handler_connect_to_start_telephony_event
       (stream->priv->stream_handler_proxy, start_telephony_event, NULL, NULL,
           (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_start_named_telephony_event
+      (stream->priv->stream_handler_proxy, start_named_telephony_event, NULL,
+          NULL, (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_start_sound_telephony_event
+      (stream->priv->stream_handler_proxy, start_sound_telephony_event, NULL,
+          NULL, (GObject*) stream, NULL);
   tp_cli_media_stream_handler_connect_to_stop_telephony_event
       (stream->priv->stream_handler_proxy, stop_telephony_event, NULL, NULL,
           (GObject*) stream, NULL);
   tp_cli_media_stream_handler_connect_to_close
       (stream->priv->stream_handler_proxy, stream_close, NULL, NULL,
           (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_set_remote_feedback_messages
+      (stream->priv->stream_handler_proxy, set_remote_feedback_messages, NULL,
+          NULL, (GObject*) stream, NULL);
+  tp_cli_media_stream_handler_connect_to_set_remote_header_extensions
+      (stream->priv->stream_handler_proxy, set_remote_header_extensions, NULL,
+          NULL, (GObject*) stream, NULL);
 
   memset (params, 0, sizeof(GParameter) * MAX_STREAM_TRANS_PARAMS);
 
@@ -995,6 +1081,22 @@ get_all_properties_cb (TpProxy *proxy,
       return;
     }
 
+  rtp_header_extensions =
+      fs_utils_get_default_rtp_header_extension_preferences (
+          GST_ELEMENT (stream->priv->fs_conference),
+          tp_media_type_to_fs (stream->priv->media_type));
+
+  if (rtp_header_extensions)
+    {
+      g_object_set (stream->priv->fs_session,
+          "rtp-header-extension-preferences", rtp_header_extensions, NULL);
+      fs_rtp_header_extension_list_destroy (rtp_header_extensions);
+    }
+
+  if (!stream->priv->local_preferences)
+    stream->priv->local_preferences = fs_utils_get_default_codec_preferences (
+        GST_ELEMENT (stream->priv->fs_conference));
+
   if (stream->priv->local_preferences)
     if (!fs_session_set_codec_preferences (stream->priv->fs_session,
             stream->priv->local_preferences,
@@ -1051,10 +1153,14 @@ async_method_callback_optional (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
     gpointer user_data,
     GObject *weak_object)
 {
-  if (error != NULL &&
-      !(error->domain == DBUS_GERROR &&
-          error->code == G_DBUS_ERROR_UNKNOWN_METHOD))
-    async_method_callback (proxy, error, user_data, weak_object);
+  if (error == NULL)
+    return;
+
+  if (g_error_matches (error, DBUS_GERROR, G_DBUS_ERROR_UNKNOWN_METHOD) ||
+      g_error_matches (error, TP_ERRORS, TP_ERROR_NOT_IMPLEMENTED))
+      return;
+
+  async_method_callback (proxy, error, user_data, weak_object);
 }
 
 static void
@@ -1517,6 +1623,23 @@ fill_fs_params (gpointer key, gpointer value, gpointer user_data)
     fs_codec_add_optional_parameter (codec, key, value);
 }
 
+static FsStreamDirection
+tpdirection_to_fsdirection (TpMediaStreamDirection dir)
+{
+  switch (dir) {
+  case TP_MEDIA_STREAM_DIRECTION_NONE:
+    return FS_DIRECTION_NONE;
+  case TP_MEDIA_STREAM_DIRECTION_SEND:
+    return FS_DIRECTION_SEND;
+  case TP_MEDIA_STREAM_DIRECTION_RECEIVE:
+    return FS_DIRECTION_RECV;
+  case TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL:
+    return FS_DIRECTION_BOTH;
+  default:
+    g_assert_not_reached ();
+  }
+}
+
 static void
 set_remote_codecs (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
     const GPtrArray *codecs,
@@ -1560,6 +1683,52 @@ set_remote_codecs (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
       params = g_value_get_boxed (g_value_array_get_nth (codec, 5));
       g_hash_table_foreach (params, fill_fs_params, fs_codec);
 
+      if (self->priv->feedback_messages)
+        {
+          GValueArray *message_props;
+
+          message_props = g_hash_table_lookup (self->priv->feedback_messages,
+              GUINT_TO_POINTER (fs_codec->id));
+
+          if (message_props)
+            {
+              GValue *val;
+              GPtrArray *messages;
+              guint j;
+
+              g_assert (G_VALUE_HOLDS_UINT (
+                      g_value_array_get_nth (message_props, 0)));
+              g_assert (G_VALUE_TYPE (
+                      g_value_array_get_nth (message_props, 1)) ==
+                  TP_ARRAY_TYPE_RTCP_FEEDBACK_MESSAGE_LIST);
+
+              val = g_value_array_get_nth (message_props, 0);
+              fs_codec->ABI.ABI.minimum_reporting_interval =
+                  g_value_get_uint (val);
+
+              val = g_value_array_get_nth (message_props, 1);
+              messages = g_value_get_boxed (val);
+
+              for (j = 0; j < messages->len; j++)
+                {
+                  GValueArray *msg = g_ptr_array_index (messages, j);
+
+                  g_assert (G_VALUE_HOLDS_STRING (
+                          g_value_array_get_nth (msg, 0)));
+                  g_assert (G_VALUE_HOLDS_STRING (
+                           g_value_array_get_nth (msg, 1)));
+                  g_assert (G_VALUE_HOLDS_STRING (
+                          g_value_array_get_nth (msg, 2)));
+
+                  fs_codec_add_feedback_parameter (fs_codec,
+                      g_value_get_string (g_value_array_get_nth (msg, 0)),
+                      g_value_get_string (g_value_array_get_nth (msg, 1)),
+                      g_value_get_string (g_value_array_get_nth (msg, 2)));
+                }
+            }
+        }
+
+
       DEBUG (self, "adding remote codec %s [%d]",
           fs_codec->encoding_name, fs_codec->id);
 
@@ -1567,6 +1736,55 @@ set_remote_codecs (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
   }
   fs_remote_codecs = g_list_reverse (fs_remote_codecs);
 
+  if (self->priv->feedback_messages)
+    {
+      g_boxed_free (TP_HASH_TYPE_RTCP_FEEDBACK_MESSAGE_MAP,
+          self->priv->feedback_messages);
+      self->priv->feedback_messages = NULL;
+    }
+
+
+  if (self->priv->header_extensions)
+    {
+      if (g_object_class_find_property (
+              G_OBJECT_GET_CLASS (self->priv->fs_stream),
+              "rtp-header-extensions"))
+        {
+          GList *hdrexts = NULL;
+
+          for (i = 0; i < self->priv->header_extensions->len; i++)
+            {
+              GValueArray *extension =
+                  g_ptr_array_index (self->priv->header_extensions, i);
+              FsRtpHeaderExtension *hdrext;
+
+              g_assert (extension->n_values >= 3);
+              g_assert (G_VALUE_HOLDS_UINT (
+                      g_value_array_get_nth (extension, 0)));
+              g_assert (G_VALUE_HOLDS_UINT (
+                      g_value_array_get_nth (extension, 1)));
+              g_assert (G_VALUE_HOLDS_STRING (
+                      g_value_array_get_nth (extension, 2)));
+
+              hdrext = fs_rtp_header_extension_new (
+                  g_value_get_uint (g_value_array_get_nth (extension, 0)),
+                  tpdirection_to_fsdirection (
+                      g_value_get_uint (g_value_array_get_nth (extension, 1))),
+                  g_value_get_string (g_value_array_get_nth (extension, 2)));
+
+              hdrexts = g_list_append (hdrexts, hdrext);
+            }
+
+          g_object_set (self->priv->fs_stream, "rtp-header-extensions",
+              hdrexts, NULL);
+
+          fs_rtp_header_extension_list_destroy (hdrexts);
+        }
+      g_boxed_free (TP_ARRAY_TYPE_RTP_HEADER_EXTENSIONS_LIST,
+          self->priv->header_extensions);
+      self->priv->header_extensions = NULL;
+    }
+
   if (!fs_stream_set_remote_codecs (self->priv->fs_stream, fs_remote_codecs,
           &error)) {
     /*
@@ -1817,13 +2035,155 @@ start_telephony_event (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
 
   DEBUG (self, "called with event %u", event);
 
+  if (self->priv->sending_telephony_event != -1)
+    {
+      WARNING (self, "start new telephony event without stopping the"
+          " previous one first");
+      if (!fs_session_stop_telephony_event (self->priv->fs_session,
+              self->priv->sending_telephony_event))
+        WARNING (self, "stopping event failed");
+    }
+
   /* this week, volume is 8, for the sake of argument... */
 
   if (!fs_session_start_telephony_event (self->priv->fs_session, event, 8,
           FS_DTMF_METHOD_AUTO))
     WARNING (self, "sending event %u failed", event);
+  self->priv->sending_telephony_event = FS_DTMF_METHOD_AUTO;
 }
 
+static gboolean
+check_codecs_for_telephone_event (TfStream *self, GList **codecs,
+    FsCodec *send_codec, guint codecid)
+{
+  GList *item = NULL;
+  gboolean found = FALSE;
+  GError *error = NULL;
+
+ again:
+
+  for (item = *codecs; item; item = item->next)
+    {
+      FsCodec *codec = item->data;
+
+      if (!g_ascii_strcasecmp (codec->encoding_name, "telephone-event") &&
+          send_codec->clock_rate == codec->clock_rate)
+        {
+          if (found)
+            {
+              *codecs = g_list_delete_link (*codecs, item);
+              goto again;
+            }
+          else if (codecid == (guint) codec->id)
+            {
+              return TRUE;            }
+          else
+            {
+              codec->id = codecid;
+            }
+        }
+    }
+
+  if (!found)
+    {
+      FsCodec *codec = fs_codec_new (codecid, "telephone-event",
+          FS_MEDIA_TYPE_AUDIO, send_codec->clock_rate);
+
+      *codecs = g_list_append (*codecs, codec);
+    }
+
+  if (!fs_stream_set_remote_codecs (self->priv->fs_stream, *codecs, &error))
+    {
+      /*
+       * Call the error method with the proper thing here
+       */
+      g_prefix_error (&error, "Codec negotiation failed for DTMF: ");
+      tf_stream_error (self, fserror_to_tperror (error), error->message);
+      g_clear_error (&error);
+    }
+
+  return FALSE;
+}
+
+static void
+start_named_telephony_event (TpMediaStreamHandler *proxy,
+    guchar event,
+    guint codecid,
+    gpointer user_data,
+    GObject *object)
+{
+  TfStream *self = TF_STREAM (object);
+  FsCodec *send_codec = NULL;
+  GList *codecs = NULL;
+  struct DtmfEvent *dtmfevent;
+
+  g_object_get (self->priv->fs_session,
+      "current-send-codec", &send_codec,
+      "codecs", &codecs,
+      NULL);
+
+
+  if (check_codecs_for_telephone_event (self, &codecs, send_codec, codecid))
+    {
+      if (self->priv->sending_telephony_event != -1)
+        {
+          WARNING (self, "start new telephony event without stopping the"
+              " previous one first");
+          if (!fs_session_stop_telephony_event (self->priv->fs_session,
+                  self->priv->sending_telephony_event))
+            WARNING (self, "stopping event failed");
+        }
+
+
+      DEBUG (self, "Sending named telephony event %d with pt %d",
+          event, codecid);
+      if (!fs_session_start_telephony_event (self->priv->fs_session,
+              event, 8, FS_DTMF_METHOD_RTP_RFC4733))
+        WARNING (self, "sending event %u failed", event);
+      self->priv->sending_telephony_event = FS_DTMF_METHOD_RTP_RFC4733;
+    }
+  else
+    {
+      DEBUG (self, "Queing named telephony event %d with pt %d",
+          event, codecid);
+      dtmfevent = g_slice_new (struct DtmfEvent);
+      dtmfevent->codec_id = codecid;
+      dtmfevent->event_id = event;
+      g_queue_push_tail (&self->priv->events_to_send, dtmfevent);
+    }
+
+  fs_codec_destroy (send_codec);
+  fs_codec_list_destroy (codecs);
+}
+
+static void
+start_sound_telephony_event (TpMediaStreamHandler *proxy, guchar event,
+    gpointer user_data, GObject *object)
+{
+  TfStream *self = TF_STREAM (object);
+
+  g_assert (self->priv->fs_session != NULL);
+
+  DEBUG (self, "called with event %u", event);
+
+  if (self->priv->sending_telephony_event != -1)
+    {
+      WARNING (self, "start new telephony event without stopping the"
+          " previous one first");
+      if (!fs_session_stop_telephony_event (self->priv->fs_session,
+              self->priv->sending_telephony_event))
+        WARNING (self, "stopping event failed");
+    }
+
+  /* this week, volume is 8, for the sake of argument... */
+
+  if (!fs_session_start_telephony_event (self->priv->fs_session, event, 8,
+          FS_DTMF_METHOD_IN_BAND))
+    WARNING (self, "sending sound event %u failed", event);
+  self->priv->sending_telephony_event = FS_DTMF_METHOD_IN_BAND;
+}
+
+
 static void
 stop_telephony_event (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
     gpointer user_data G_GNUC_UNUSED,
@@ -1835,6 +2195,11 @@ stop_telephony_event (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
 
   DEBUG (self, "called");
 
+  if (self->priv->sending_telephony_event == -1)
+      WARNING (self, "Trying to stop telephony event without having started"
+          " one");
+  self->priv->sending_telephony_event = -1;
+
   if (!fs_session_stop_telephony_event (self->priv->fs_session,
           FS_DTMF_METHOD_AUTO))
     WARNING (self, "stopping event failed");
@@ -1867,6 +2232,38 @@ stream_close (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
   tf_stream_shutdown (self);
 }
 
+
+static void
+set_remote_feedback_messages (TpMediaStreamHandler *proxy,
+    GHashTable *messages, gpointer user_data, GObject *object)
+{
+  TfStream *self = TF_STREAM (object);
+
+  if (self->priv->feedback_messages)
+    g_boxed_free (TP_HASH_TYPE_RTCP_FEEDBACK_MESSAGE_MAP,
+        self->priv->feedback_messages);
+
+  self->priv->feedback_messages =
+      g_boxed_copy (TP_HASH_TYPE_RTCP_FEEDBACK_MESSAGE_MAP, messages);
+}
+
+
+static void
+set_remote_header_extensions (TpMediaStreamHandler *proxy,
+    const GPtrArray *header_extensions, gpointer user_data, GObject *object)
+{
+  TfStream *self = TF_STREAM (object);
+
+  if (self->priv->header_extensions)
+    g_boxed_free (TP_ARRAY_TYPE_RTP_HEADER_EXTENSIONS_LIST,
+        self->priv->header_extensions);
+
+  self->priv->header_extensions =
+      g_boxed_copy (TP_ARRAY_TYPE_RTP_HEADER_EXTENSIONS_LIST,
+          header_extensions);
+}
+
+
 static void
 cb_fs_recv_codecs_changed (TfStream *self,
     GList *codecs)
@@ -1966,6 +2363,72 @@ invalidated_cb (TpMediaStreamHandler *proxy G_GNUC_UNUSED,
   tf_stream_shutdown (stream);
 }
 
+static void
+cb_fs_send_codec_changed (TfStream *self,
+    FsCodec *send_codec,
+    GList *secondary_codecs)
+{
+  GList *item;
+  gint last_event_id = -1;
+  struct DtmfEvent *dtmfevent;
+
+  while ((dtmfevent = g_queue_peek_head (&self->priv->events_to_send)))
+    {
+      if (dtmfevent->codec_id != last_event_id)
+        {
+          last_event_id = -1;
+          for (item = secondary_codecs; item; item = item->next)
+            {
+              FsCodec *codec = item->data;
+
+              if (!g_ascii_strcasecmp (codec->encoding_name, "telephone-event")
+                  && codec->id == dtmfevent->codec_id)
+                {
+                  last_event_id = codec->id;
+                  goto have_id;
+                }
+            }
+          if (dtmfevent->codec_id != last_event_id)
+            {
+              GList *codecs = NULL;
+
+              g_object_get (self->priv->fs_session, "codecs", &codecs, NULL);
+
+              DEBUG (self, "Still do not have the right PT for telephony"
+                  " events, trying to force it again");
+              if (check_codecs_for_telephone_event (self, &codecs, send_codec,
+                      dtmfevent->codec_id))
+                WARNING (self, "Did not have the right pt in the secondary"
+                    " codecs, but it was in the codec list. Ignoring for now");
+              fs_codec_list_destroy (codecs);
+              return;
+            }
+        }
+
+    have_id:
+      DEBUG (self, "Sending queued event %d with pt %d", dtmfevent->event_id,
+          dtmfevent->codec_id);
+      dtmfevent = g_queue_pop_head (&self->priv->events_to_send);
+      if (self->priv->sending_telephony_event != -1)
+        {
+          WARNING (self, "start new telephony event without stopping the"
+              " previous one first");
+          if (!fs_session_stop_telephony_event (self->priv->fs_session,
+                  self->priv->sending_telephony_event))
+            WARNING (self, "stopping event failed");
+        }
+      self->priv->sending_telephony_event = -1;
+
+      if (!fs_session_start_telephony_event (self->priv->fs_session,
+              dtmfevent->event_id, 8, FS_DTMF_METHOD_RTP_RFC4733))
+        WARNING (self, "sending event %u failed", dtmfevent->event_id);
+      fs_session_stop_telephony_event (self->priv->fs_session,
+          FS_DTMF_METHOD_RTP_RFC4733);
+
+      g_slice_free (struct DtmfEvent, dtmfevent);
+    }
+}
+
 /**
  * tf_stream_error:
  * @self: a #TfStream
@@ -2140,12 +2603,16 @@ _tf_stream_bus_message (TfStream *stream,
       DEBUG (stream, "Codecs changed");
 
       _tf_stream_try_sending_codecs (stream);
+
+      return TRUE;
     }
   else if (gst_structure_has_name (s, "farsight-send-codec-changed"))
     {
       FsSession *fssession;
       const GValue *value;
       FsCodec *codec = NULL;
+      GList *secondary_codecs = NULL;
+      FsCodec *objcodec = NULL;
 
       value = gst_structure_get_value (s, "session");
       fssession = g_value_get_object (value);
@@ -2153,15 +2620,27 @@ _tf_stream_bus_message (TfStream *stream,
       if (fssession != stream->priv->fs_session)
         return FALSE;
 
-      g_object_get (fssession, "current-send-codec", &codec, NULL);
+      value = gst_structure_get_value (s, "codec");
+      codec = g_value_get_boxed (value);
 
-      if (codec)
+      g_object_get (fssession, "current-send-codec", &objcodec, NULL);
+      if (!fs_codec_are_equal (objcodec, codec))
         {
-          DEBUG (stream, "Send codec changed: " FS_CODEC_FORMAT,
-              FS_CODEC_ARGS (codec));
-
-          fs_codec_destroy (codec);
+          fs_codec_destroy (objcodec);
+          return TRUE;
         }
+      fs_codec_destroy (objcodec);
+
+      value = gst_structure_get_value (s, "secondary-codecs");
+      secondary_codecs = g_value_get_boxed (value);
+
+
+      DEBUG (stream, "Send codec changed: " FS_CODEC_FORMAT,
+          FS_CODEC_ARGS (codec));
+
+      cb_fs_send_codec_changed (stream, codec, secondary_codecs);
+
+      return TRUE;
     }
   else if (gst_structure_has_name (s, "farsight-component-state-changed"))
     {
@@ -2182,6 +2661,7 @@ _tf_stream_bus_message (TfStream *stream,
         return TRUE;
 
       cb_fs_component_state_changed (stream, component, fsstate);
+      return TRUE;
     }
 
   return FALSE;
@@ -2266,6 +2746,119 @@ _tf_stream_new (gpointer channel,
   return self;
 }
 
+static GHashTable *
+fs_codecs_to_feedback_messages (GList *fscodecs)
+{
+  GList *item;
+  GHashTable *feedback_messages = g_hash_table_new_full (g_direct_hash,
+      g_direct_equal, NULL, (GDestroyNotify) g_value_array_free);
+
+  for (item = fscodecs; item; item = item->next)
+    {
+      FsCodec *fs_codec = item->data;
+
+      if (fs_codec->ABI.ABI.minimum_reporting_interval != G_MAXUINT ||
+          fs_codec->ABI.ABI.feedback_params)
+        {
+          GValueArray *codec = g_value_array_new (2);
+          GPtrArray *messages = g_ptr_array_new ();
+          GValue *val;
+          GList *item2;
+
+          for (item2 = fs_codec->ABI.ABI.feedback_params;
+               item2;
+               item2 = item2->next)
+            {
+              FsFeedbackParameter *p = item2->data;
+              GValueArray *message = g_value_array_new (3);
+              GValue *val2;
+
+              g_value_array_insert (message, 0, NULL);
+              val2 = g_value_array_get_nth (message, 0);
+              g_value_init (val2, G_TYPE_STRING);
+              g_value_set_string (val2, p->type);
+
+              g_value_array_insert (message, 1, NULL);
+              val2 = g_value_array_get_nth (message, 1);
+              g_value_init (val2, G_TYPE_STRING);
+              g_value_set_string (val2, p->subtype);
+
+              g_value_array_insert (message, 2, NULL);
+              val2 = g_value_array_get_nth (message, 2);
+              g_value_init (val2, G_TYPE_STRING);
+              g_value_set_string (val2, p->extra_params);
+
+              g_ptr_array_add (messages, message);
+            }
+
+          g_value_array_insert (codec, 0, NULL);
+          val = g_value_array_get_nth (codec, 0);
+          g_value_init (val, G_TYPE_UINT);
+          g_value_set_uint (val, fs_codec->ABI.ABI.minimum_reporting_interval);
+
+          g_value_array_insert (codec, 1, NULL);
+          val = g_value_array_get_nth (codec, 1);
+          g_value_init (val, TP_ARRAY_TYPE_RTCP_FEEDBACK_MESSAGE_LIST);
+          g_value_take_boxed (val, messages);
+
+          g_hash_table_insert (feedback_messages,
+              GUINT_TO_POINTER (fs_codec->id), codec);
+        }
+    }
+
+  return feedback_messages;
+}
+
+
+static TpMediaStreamDirection
+fsdirection_to_tpdirection (FsStreamDirection dir)
+{
+  switch (dir) {
+  case FS_DIRECTION_NONE:
+    return TP_MEDIA_STREAM_DIRECTION_NONE;
+  case FS_DIRECTION_SEND:
+    return TP_MEDIA_STREAM_DIRECTION_SEND;
+  case FS_DIRECTION_RECV:
+    return TP_MEDIA_STREAM_DIRECTION_RECEIVE;
+  case FS_DIRECTION_BOTH:
+    return TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL;
+  default:
+    g_assert_not_reached ();
+  }
+}
+
+
+static GPtrArray *
+_tf_stream_get_header_extensions (TfStream *stream)
+{
+  GPtrArray *extensions = g_ptr_array_new ();
+  GList *hdrexts;
+  GList *item;
+
+  if (!g_object_class_find_property (
+          G_OBJECT_GET_CLASS (stream->priv->fs_session),
+          "rtp-header-extensions"))
+    return extensions;
+
+  g_object_get (stream->priv->fs_session,
+      "rtp-header-extensions", &hdrexts, NULL);
+
+  for (item = hdrexts; item; item = item->next)
+    {
+      FsRtpHeaderExtension *hdrext = item->data;
+
+      g_ptr_array_add (extensions,
+          tp_value_array_build (4,
+              G_TYPE_UINT, hdrext->id,
+              G_TYPE_UINT, fsdirection_to_tpdirection (hdrext->direction),
+              G_TYPE_STRING, hdrext->uri,
+              G_TYPE_STRING, "",
+              G_TYPE_INVALID));
+    }
+
+  return extensions;
+}
+
 void
 _tf_stream_try_sending_codecs (TfStream *stream)
 {
@@ -2273,6 +2866,13 @@ _tf_stream_try_sending_codecs (TfStream *stream)
   GList *fscodecs = NULL;
   GList *item = NULL;
   GPtrArray *tpcodecs = NULL;
+  GHashTable *feedback_messages = NULL;
+  GPtrArray *header_extensions = NULL;
+  gboolean sent = FALSE;
+  GList *resend_codecs = NULL;
+
+  if (stream->priv->block_ready)
+    return;
 
   DEBUG (stream, "called (send_local:%d send_supported:%d)",
       stream->priv->send_local_codecs, stream->priv->send_supported_codecs);
@@ -2298,26 +2898,53 @@ _tf_stream_try_sending_codecs (TfStream *stream)
   if (stream->priv->send_local_codecs)
     {
       tpcodecs = fs_codecs_to_tp (stream, fscodecs);
+      feedback_messages = fs_codecs_to_feedback_messages (fscodecs);
+      header_extensions = _tf_stream_get_header_extensions (stream);
 
       DEBUG (stream, "calling MediaStreamHandler::Ready");
+      tp_cli_media_stream_handler_call_supported_feedback_messages (
+          stream->priv->stream_handler_proxy,
+          -1, feedback_messages, async_method_callback_optional,
+          "Media.StreamHandler::SupportedFeedbackMessages for Ready",
+          NULL, (GObject *) stream);
+      tp_cli_media_stream_handler_call_supported_header_extensions (
+          stream->priv->stream_handler_proxy,
+          -1, header_extensions, async_method_callback_optional,
+          "Media.StreamHandler::SupportedHeaderExtensions for Ready",
+          NULL, (GObject *) stream);
       tp_cli_media_stream_handler_call_ready (
           stream->priv->stream_handler_proxy,
           -1, tpcodecs, async_method_callback, "Media.StreamHandler::Ready",
           NULL, (GObject *) stream);
       stream->priv->send_local_codecs = FALSE;
+      stream->priv->ready_called = TRUE;
+      sent = TRUE;
       goto out;
     }
 
   if (stream->priv->send_supported_codecs)
     {
       tpcodecs = fs_codecs_to_tp (stream, fscodecs);
+      feedback_messages = fs_codecs_to_feedback_messages (fscodecs);
+      header_extensions = _tf_stream_get_header_extensions (stream);
 
       DEBUG (stream, "calling MediaStreamHandler::SupportedCodecs");
+      tp_cli_media_stream_handler_call_supported_feedback_messages (
+          stream->priv->stream_handler_proxy,
+          -1, feedback_messages, async_method_callback_optional,
+          "Media.StreamHandler::SupportedFeedbackMessages for SupportedCodecs",
+          NULL, (GObject *) stream);
+      tp_cli_media_stream_handler_call_supported_header_extensions (
+          stream->priv->stream_handler_proxy,
+          -1, header_extensions, async_method_callback_optional,
+          "Media.StreamHandler::SupportedHeaderExtensions for SupportedCodecs",
+          NULL, (GObject *) stream);
       tp_cli_media_stream_handler_call_supported_codecs (
           stream->priv->stream_handler_proxy,
           -1, tpcodecs, async_method_callback,
           "Media.StreamHandler::SupportedCodecs", NULL, (GObject *) stream);
       stream->priv->send_supported_codecs = FALSE;
+      sent = TRUE;
 
       /* Fallthrough to potentially call CodecsUpdated as CMs assume
        * SupportedCodecs will only give the intersection of the already sent
@@ -2328,22 +2955,50 @@ _tf_stream_try_sending_codecs (TfStream *stream)
   /* Only send updates if there was something to update (iotw we sent codecs
    * before) or our list changed */
   if (stream->priv->last_sent_codecs != NULL
-      && !fs_codec_list_are_equal (fscodecs, stream->priv->last_sent_codecs))
+      && (resend_codecs =
+          fs_session_codecs_need_resend (stream->priv->fs_session,
+              stream->priv->last_sent_codecs, fscodecs)) != NULL)
     {
-      tpcodecs = fs_codecs_to_tp (stream, fscodecs);
+      fs_codec_list_destroy (resend_codecs);
+
+      if (!tpcodecs)
+        tpcodecs = fs_codecs_to_tp (stream, fscodecs);
+      if (!feedback_messages)
+        feedback_messages = fs_codecs_to_feedback_messages (fscodecs);
+      if (!header_extensions)
+        header_extensions = _tf_stream_get_header_extensions (stream);
+
 
       DEBUG (stream, "calling MediaStreamHandler::CodecsUpdated");
+      tp_cli_media_stream_handler_call_supported_feedback_messages (
+          stream->priv->stream_handler_proxy,
+          -1, feedback_messages, async_method_callback_optional,
+          "Media.StreamHandler::SupportedFeedbackMessages for CodecsUpdated",
+          NULL, (GObject *) stream);
+      tp_cli_media_stream_handler_call_supported_header_extensions (
+          stream->priv->stream_handler_proxy,
+          -1, header_extensions, async_method_callback_optional,
+          "Media.StreamHandler::SupportedHeaderExtensions for CodecsUpdated",
+          NULL, (GObject *) stream);
       tp_cli_media_stream_handler_call_codecs_updated (
           stream->priv->stream_handler_proxy,
           -1, tpcodecs, async_method_callback,
           "Media.StreamHandler::CodecsUpdated", NULL, (GObject *) stream);
+      sent = TRUE;
     }
 
 out:
   if (tpcodecs)
     g_boxed_free (TP_ARRAY_TYPE_MEDIA_STREAM_HANDLER_CODEC_LIST, tpcodecs);
-  fs_codec_list_destroy (stream->priv->last_sent_codecs);
-  stream->priv->last_sent_codecs = fscodecs;
+  if (feedback_messages)
+    g_boxed_free (TP_HASH_TYPE_RTCP_FEEDBACK_MESSAGE_MAP, feedback_messages);
+  if (header_extensions)
+    g_boxed_free (TP_ARRAY_TYPE_RTP_HEADER_EXTENSIONS_LIST, header_extensions);
+  if (sent)
+    {
+      fs_codec_list_destroy (stream->priv->last_sent_codecs);
+      stream->priv->last_sent_codecs = fscodecs;
+    }
 }
 
 /**
